const webpack = require('webpack');
const MiniCssExtractPlugin = require('mini-css-extract-plugin');
const HappyPack = require('happypack');
const ESLintPlugin = require('eslint-webpack-plugin');
const ProgressBarPlugin = require('progress-bar-webpack-plugin');
const getCSSModuleLocalIdent = require('./getCSSModuleLocalIdent');
const { getProjectPath, isExistsSyncFile } = require('./utils');

const appWebpackConfigPath = getProjectPath('/config/app.webpack.js');

const happyThreadPool = HappyPack.ThreadPool({ size: 5 });
const sassRegex = /\.(scss|sass)$/;
const sassModuleRegex = /\.module\.(scss|sass)$/;

function getStyleLoaders({ isProd, isServer, isHotServer, isSass, moduleOption = {} }) {
  const result = [
    {
      loader: 'css-loader',
      options: {
        importLoaders: 1,
        ...moduleOption
      }
    },
    {
      loader: 'postcss-loader',
      options: {
        ident: 'postcss',
        plugins: () => [
          require('postcss-flexbugs-fixes'),
          require('postcss-preset-env')({
            autoprefixer: {
              flexbox: 'no-2009',
              remove: false,
              grid: true
            },
            stage: 3
          })
        ],
        sourceMap: false
      }
    }
  ];

  if (isSass) {
    result.push({
      loader: 'sass-loader'
    });
  }

  if ((isProd || isHotServer) && !isServer) {
    result.unshift({
      loader: MiniCssExtractPlugin.loader,
      options: {
        hmr: false
      }
    });
  }

  if (!isProd && !isHotServer) {
    result.unshift({
      loader: 'style-loader'
    });
  }

  return result;
}

module.exports = (appConfig, extendEnv, { isProd, isServer, isHotServer, isSetCommonAlias }) => {
  const result = {
    // 原始源代码（仅限行）定位到原始代码问题，比source-map 快，但是没有列映射(column mapping)的 source-map
    devtool: isProd ? 'source-map' : 'cheap-module-source-map',
    // 会将 process.env.NODE_ENV 的值设为 development, 告知 webpack 使用相应模式的内置优化
    mode: isProd ? 'production' : 'development',
    // 打包的出口
    output: {
      publicPath: '/'
    },
    // 配置各种loader
    module: {
      // 警告替换为Error
      strictExportPresence: true,
      rules: [
        {
          oneOf: [
            {
              test: /\.(js|ts|tsx)$/,
              exclude: /node_modules/,
              use: 'happypack/loader?id=jsbabel'
            },
            {
              test: sassRegex,
              exclude: sassModuleRegex,
              use: getStyleLoaders({
                isProd,
                isServer,
                isHotServer,
                isSass: true
              })
            },
            {
              test: sassModuleRegex,
              use: getStyleLoaders({
                isProd,
                isServer,
                isHotServer,
                isSass: true,
                moduleOption: {
                  getLocalIdent: getCSSModuleLocalIdent,
                  importLoaders: 2,
                  modules: true
                }
              })
            },
            {
              test: /\.(css)$/,
              use: getStyleLoaders({
                isProd,
                isServer,
                isHotServer
              })
            },
            {
              test: [/\.jpe?g$/, /\.png$/, /\.bmp$/, /\.gif$/],
              loader: 'url-loader',
              options: {
                limit: 2048,
                name: `static/[name].[contenthash:8].[ext]?v=${extendEnv.version}`
              }
            }
          ]
        }
      ]
    },
    resolve: {
      // 设置路径别名
      alias: appConfig.alias,
      // 文件后缀自动补全
      extensions: ['.tsx', '.ts', '.js', 'scss', 'css']
    },
    plugins: [
      new webpack.DefinePlugin(extendEnv.stringified),
      new ESLintPlugin({
        failOnError: false,
        extensions: ['ts', 'tsx', 'js']
      }),
      new HappyPack({
        id: 'jsbabel',
        verbose: false,
        threadPool: happyThreadPool,
        loaders: [
          {
            loader: 'babel-loader',
            options: {
              cacheDirectory: true,
              presets: [
                'react-app',
                '@babel/preset-env',
                '@babel/preset-react',
                '@babel/typescript'
              ],
              plugins: [
                '@loadable/babel-plugin',
                '@babel/plugin-transform-object-assign',
                '@babel/plugin-transform-shorthand-properties',
                '@babel/plugin-transform-runtime',
                [
                  '@babel/plugin-proposal-decorators',
                  {
                    legacy: true
                  }
                ],
                [
                  '@babel/plugin-proposal-class-properties',
                  {
                    loose: false
                  }
                ]
              ]
            }
          }
        ]
      }),
      new ProgressBarPlugin({
        // eslint-disable-next-line no-useless-concat
        format: 'build [:bar] ' + ':percent' + ' (:elapsed seconds)',
        clear: false
      })
    ]
  };
  if (isProd && !isServer) {
    result.optimization = {
      minimize: true
    };
  }

  if (isExistsSyncFile(appWebpackConfigPath)) {
    const extendConfig = require(appWebpackConfigPath);
    if (typeof extendConfig !== 'function') {
      throw Error(
        `Extend webpack config file only functions are supported and return extend result! \n current type is ${typeof extendConfig} \n filename is ${appWebpackConfigPath}`
      );
    }
    const extendResult = extendConfig(result, appConfig, extendEnv, {
      isProd,
      isServer,
      isHotServer,
      isSetCommonAlias
    });
    if (!extendResult) {
      throw Error(
        `Return extend result! \n current extendResult is ${extendResult} \n filename is ${appWebpackConfigPath}`
      );
    }
    return extendResult;
  }

  return result;
};
